package ru.dbotthepony.kommons.io

import java.io.DataInput
import java.io.InputStream
import java.io.RandomAccessFile
import java.math.BigDecimal
import java.math.BigInteger
import java.util.function.IntSupplier

fun InputStream.readBigDecimal(): BigDecimal {
	val scale = readInt()
	val size = readVarInt()
	require(size >= 0) { "Negative payload size: $size" }
	val bytes = ByteArray(size)
	read(bytes)
	return BigDecimal(BigInteger(bytes), scale)
}

fun <S : InputStream, V, C : MutableCollection<V>> S.readCollection(reader: S.() -> V, factory: (Int) -> C): C {
	val size = readVarInt()
	val collection = factory.invoke(size)

	for (i in 0 until size) {
		collection.add(reader(this))
	}

	return collection
}

fun <S : InputStream, V> S.readCollection(reader: S.() -> V) = readCollection(reader, ::ArrayList)

fun InputStream.readInt(): Int {
	if (this is DataInput) {
		return readInt()
	}

	return (read() shl 24) or (read() shl 16) or (read() shl 8) or read()
}

fun InputStream.readLong(): Long {
	if (this is DataInput) {
		return readLong()
	}

	return (read().toLong() shl 48) or
		(read().toLong() shl 40) or
		(read().toLong() shl 32) or
		(read().toLong() shl 24) or
		(read().toLong() shl 16) or
		(read().toLong() shl 8) or
		read().toLong()
}

fun InputStream.readFloat() = Float.fromBits(readInt())
fun InputStream.readDouble() = Double.fromBits(readLong())

data class VarIntReadResult(val value: Int, val cells: Int)
data class VarLongReadResult(val value: Long, val cells: Int)

private fun readVarLongInfo(supplier: IntSupplier): VarLongReadResult {
	var result = 0L
	var read = supplier.asInt
	var i = 1

	while (true) {
		result = (result shl 7) or (read.toLong() and 0x7F)

		if (read and 0x80 == 0)
			break

		read = supplier.asInt
		i++
	}

	return VarLongReadResult(result, i)
}

private fun readVarIntInfo(supplier: IntSupplier): VarIntReadResult {
	val read = readVarLongInfo(supplier)
	return VarIntReadResult(read.value.toInt(), read.cells)
}

private fun readVarInt(supplier: IntSupplier): Int {
	return readVarIntInfo(supplier).value
}

private fun readVarLong(supplier: IntSupplier): Long {
	return readVarLongInfo(supplier).value
}

private fun Int.fromSignedVar(): Int {
	val sign = this and 1

	if (sign == 0)
		return this ushr 1
	else
		return -(this ushr 1) - 1
}

private fun VarIntReadResult.fromSignedVar(): VarIntReadResult {
	val sign = this.value and 1

	if (sign == 0)
		return VarIntReadResult(this.value ushr 1, this.cells)
	else
		return VarIntReadResult(-(this.value ushr 1) - 1, this.cells)
}

private fun Long.fromSignedVar(): Long {
	val sign = this and 1

	if (sign == 0L)
		return this ushr 1
	else
		return -(this ushr 1) - 1L
}

private fun VarLongReadResult.fromSignedVar(): VarLongReadResult {
	val sign = this.value and 1

	if (sign == 0L)
		return VarLongReadResult(this.value ushr 1, this.cells)
	else
		return VarLongReadResult(-(this.value ushr 1) - 1, this.cells)
}

fun RandomAccessFile.readVarLong() = readVarLong(::readUnsignedByte)
fun RandomAccessFile.readVarInt() = readVarInt(::readUnsignedByte)
fun RandomAccessFile.readVarLongInfo() = readVarLongInfo(::readUnsignedByte)
fun RandomAccessFile.readVarIntInfo() = readVarIntInfo(::readUnsignedByte)
fun InputStream.readVarLong() = readVarLong(::read)
fun InputStream.readVarInt() = readVarInt(::read)
fun InputStream.readVarLongInfo() = readVarLongInfo(::read)
fun InputStream.readVarIntInfo() = readVarIntInfo(::read)

fun RandomAccessFile.readSignedVarLong() = readVarLong(::readUnsignedByte).fromSignedVar()
fun RandomAccessFile.readSignedVarInt() = readVarInt(::readUnsignedByte).fromSignedVar()
fun RandomAccessFile.readSignedVarLongInfo() = readVarLongInfo(::readUnsignedByte).fromSignedVar()
fun RandomAccessFile.readSignedVarIntInfo() = readVarIntInfo(::readUnsignedByte).fromSignedVar()
fun InputStream.readSignedVarLong() = readVarLong(::read).fromSignedVar()
fun InputStream.readSignedVarInt() = readVarInt(::read).fromSignedVar()
fun InputStream.readSignedVarLongInfo() = readVarLongInfo(::read).fromSignedVar()
fun InputStream.readSignedVarIntInfo() = readVarIntInfo(::read).fromSignedVar()

fun InputStream.readBinaryString(): String {
	val size = readVarInt()
	require(size >= 0) { "Negative payload size: $size" }
	val bytes = ByteArray(size)
	read(bytes)
	return bytes.decodeToString()
}
