package ru.dbotthepony.kommons.io

import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.*
import kotlin.reflect.KClass

class StreamCodec<V>(
	private val reader: (stream: DataInputStream) -> V,
	private val writer: (stream: DataOutputStream, value: V) -> Unit,
	private val copier: ((value: V) -> V) = { it },
	private val comparator: ((a: V, b: V) -> Boolean) = { a, b -> a == b }
) : IStreamCodec<V> {
	constructor(
		reader: (stream: DataInputStream) -> V,
		payloadSize: Long,
		writer: (stream: DataOutputStream, value: V) -> Unit,
		copier: ((value: V) -> V) = { it },
		comparator: ((a: V, b: V) -> Boolean) = { a, b -> a == b }
	) : this({ stream -> reader.invoke(stream) }, writer, copier, comparator)

	val nullable = object : IStreamCodec<V?> {
		override fun read(stream: DataInputStream): V? {
			return if (stream.read() == 0) null else reader.invoke(stream)
		}

		override fun write(stream: DataOutputStream, value: V?) {
			if (value === null) stream.write(0) else {
				stream.write(1)
				writer.invoke(stream, value)
			}
		}

		override fun copy(value: V?): V? {
			return if (value === null) null else copier.invoke(value)
		}

		override fun compare(a: V?, b: V?): Boolean {
			if (a === null && b === null) return true
			if (a === null || b === null) return false
			return comparator.invoke(a, b)
		}
	}

	override fun read(stream: DataInputStream): V {
		return reader.invoke(stream)
	}

	override fun write(stream: DataOutputStream, value: V) {
		writer.invoke(stream, value)
	}

	override fun copy(value: V): V {
		return copier.invoke(value)
	}

	override fun compare(a: V, b: V): Boolean {
		return comparator.invoke(a, b)
	}
}

val NullValueCodec = StreamCodec({ _ -> null }, { _, _ -> })
val BooleanValueCodec = StreamCodec(DataInputStream::readBoolean, 1L, DataOutputStream::writeBoolean) { a, b -> a == b }
val ByteValueCodec = StreamCodec(DataInputStream::readByte, 1L, { s, v -> s.writeByte(v.toInt()) }) { a, b -> a == b }
val ShortValueCodec = StreamCodec(DataInputStream::readShort, 2L, { s, v -> s.writeShort(v.toInt()) }) { a, b -> a == b }
val IntValueCodec = StreamCodec(DataInputStream::readInt, 4L, DataOutputStream::writeInt) { a, b -> a == b }
val LongValueCodec = StreamCodec(DataInputStream::readLong, 8L, DataOutputStream::writeLong) { a, b -> a == b }
val FloatValueCodec = StreamCodec(DataInputStream::readFloat, 4L, DataOutputStream::writeFloat) { a, b -> a == b }
val DoubleValueCodec = StreamCodec(DataInputStream::readDouble, 8L, DataOutputStream::writeDouble) { a, b -> a == b }
val BigDecimalValueCodec = StreamCodec(DataInputStream::readBigDecimal, DataOutputStream::writeBigDecimal)
val UUIDValueCodec = StreamCodec({ s -> UUID(s.readLong(), s.readLong()) }, { s, v -> s.writeLong(v.mostSignificantBits); s.writeLong(v.leastSignificantBits) })
val VarIntValueCodec = StreamCodec(DataInputStream::readSignedVarInt, DataOutputStream::writeSignedVarInt) { a, b -> a == b }
val VarLongValueCodec = StreamCodec(DataInputStream::readSignedVarLong, DataOutputStream::writeSignedVarLong) { a, b -> a == b }
val BinaryStringCodec = StreamCodec(DataInputStream::readBinaryString, DataOutputStream::writeBinaryString)

fun <E : Enum<E>> Class<E>.codec() = EnumValueCodec(this)
fun <E : Enum<E>> KClass<E>.codec() = EnumValueCodec(this.java)
